%Authors: 	Martin Holler and Kristian Bredies
%           	University of Graz
%Contact	martin.holler@uni-graz.at
%Usage permited only for academic purposes
%NOTE: This is NON-OPTIMIZED matlab code.


%Function for TGV-based JPEG decompression of color images with optional
%zooming.

%Restrictions:
%   Image dimension must be multiples of 8
%   Requires Independent JPEG Group's JPEG runtime library and headers (tested with version 6.2)


%Usage:

%[g1,g2,tgv,gap,par_out] = tgv_jpeg(img_name,par_in);

    %This decompresses the JPEG File or simulates jpeg decompression on an uncompressed image
    %Input parameters are:
    %   img_name  Name of .jpg file or uncompressed image file (e.g. .png, .tiff)
    %   par_in  n x 2 cell of parameter: 
    %       par_in{i,1}     parameter name,
    %       par_in{i,2}     parameter value

    %Output parameters are:
    %   g1      Standard decompression
    %   g2      TGV-based decompression
    %   tgv     TGV value of solution
    %   gap     Duality gap value of solution
    %   par_out Parameter cell output

    
%Example usage:
    
    %Decompress real jpeg file and set stopping rule
    %[g1,g2,tgv,gap,par_out] = tgv_jpeg('bird_compressed.jpg',{'stop_rule','iteration';'stop_par',500});
    
%Admisible parameter values are

    %{'sig',par}    Dual steplength, par is a positive real number
    %{'tau',par}    Primal steplength, par is a positive real number
    %{'stop_rule',par}  Stopping rule
    %       par = 'iteration' means maximum iteration number
    %       par = 'gap' means duality gap
    %{'stop_par',par}   Stopping parameter
    %       par = number of iterations if 'stop_rule' = 'iteration'
    %       par = duality gap tolerance if 'stop_rule' = 'gap'
    %{'rate',par}   Compression rate, only applies if input is an original
    %               uncompressed image array and not a string.
    %{'alph0',par}  par is a positive real number, parameter for TGV
    %{'alph1',par}  par is a positive real number, parameter for TGV
    %{'zoom_fac',par}   Zooming factor. Par is a natural number. If par > 1
    %                   the decompressed image is magnified by factor par
    %{'subsample_mode',par} Subsampling mode. Applies if color components
    %                       are subsampled or if zooming is desired
    %       par = 'dct' Upsamling is achieved by filling unknown DCT data
    %       par = 'explicit' Upsampling by explicit pixel repetition
    
    %{'gamma',par}  par is a real number, parameter for duality gap
    %{'p',par}      par is a real number larger than 1 or infty, 
    %               parameter for duality gap


function [g1,g2,tgv,gap,par_out] = tgv_jpeg(img_name,par_in)


%Setpaths
addpath standard_functions  %Add path to standard functions
addpath data_handling       %Add path to data fidelity functions


%% Set Parameter

    %Set standard parameter------------------------------------------------
    
    
    %Steplength
    sig = sqrt(1/12);
    tau = sqrt(1/12);
    
    %Stopping rule   
    stop_rule = 'iteration'; %Options: 'gap','iteration'
    stop_par = 100;
    
    %Compression rate for internal compression
    rate = 3;
    
    %TGV weights
    alph0 = sqrt(2);
    alph1 = 1;
    
    %Zooming parameter
    zoom_fac = 1;
    subsample_mode = 'dct'; %Options: 'dct', 'explicit'
    
    %Internal stoppin rule parameter
    gamma = 1.001;
    p = 2;
    
    %Directives
    plot = 0;
    update_rate = 500;
    display_images = 0;
    get_gap = 0;


    %% Read parameter##########################################################
    %Input: par_in
    %Generate list of parameters
    vars = whos;
    for l=1:size(vars,1)
        par_list{l,1} = vars(l).name;
    end

    %Set parameters according to list
    for l=1:size(par_in,1);
        valid = false;
        for j=1:size(par_list,1); if strcmp(par_in{l,1},par_list{j,1})
                valid = true;
                eval([par_in{l,1},'=','par_in{l,2}',';']);
        end; end
        if valid == false; warning(['Unexpected parameter at ',num2str(l)]); end
    end
    %##########################################################################
    
    
    
%%Read image
    if strcmp(img_name(end-2:end),'jpg') || strcmp(img_name(end-3:end),'jpeg') %If input file is jpeg image
        
        [coeff_data,quant_data,col_info] = read_jpeg_data(img_name);
        for ii=1:length(coeff_data)
            coeff_data{ii} = double(coeff_data{ii}');
            quant_data{ii} = double(quant_data{ii}');
        end
        
                
    else %If input file is uncompressed image file
        
        display('Regarding input as uncompressed image and simulating jpeg compression.');
        image = imread(img_name);
        
        col_info = 0;%Default value
        if size(image,3)==3
        	image = double(rgb2ycbcr(image));
        	col_info=3;
        end
        
        for ii=1:size(image,3)
            [coeff_data{ii},quant_data{ii}] = jpeg_enc(image(:,:,ii),rate);
        end
        
        
    end

    %Number of color components
    ncol = length(coeff_data);
    
    
    %Set component size
    n = zeros(1,ncol);
    m = zeros(1,ncol);
    for ii=1:ncol
        n(ii) = size(coeff_data{ii},1);
        m(ii) = size(coeff_data{ii},2);
    end
    
    
    N = zoom_fac*n(1); %Set disired image size
    M = zoom_fac*m(1);

    %If dct subsampling is used, the data is assumed to have the same size
    %as the disired image size by filling unknown intervals by \R
    if strcmp(subsample_mode,'dct')
        dn = repmat(N,[1,ncol]);
        dm = repmat(M,[1,ncol]);
    %If we use the subsamping operator, the data is assumed to have the actuall
    %data size, no unbounded intervals occure
    elseif strcmp(subsample_mode,'explicit')
        dn = n; dm = m;
    end


    %Set zooming factors for dct zooming
    facn = dn./n;
    facm = dm./m;


    %Display image Properties
    display( '-------------------------------------');
    display(['Actual image dimensions:    ',num2str(n(1)),'X',num2str(m(1))]) %Actual image size
    display(['Disired image dimensions:   ',num2str(N),'X',num2str(M)]) %Actual image size
    %Storage demand
    display(['Size of uncompressed high resolution image: ',num2str(3*N*M)])
    d_size = 0;
    for ii=1:ncol
        d_size = d_size + nnz(coeff_data{ii});
    end
    display(['Size of compressed image:   ',num2str(d_size)])

    %Dequantization of data----------------------------------------------------
    for ii=1:ncol
        for i=1:(n(ii)/8)
            for j=1:(m(ii)/8)
                     coeff_data{ii}(8*(i-1)+1:8*i,8*(j-1)+1:8*j)=coeff_data{ii}(8*(i-1)+1:8*i,8*(j-1)+1:8*j).*quant_data{ii}; %dequantization of data
            end
        end
    end
    

    %Standart decompression at original size-------------------------------
    g1=zeros(n(1),m(1),ncol); %Prepare space for standard decompresion


    %Calculation of direct inverse DCT
    for ii=1:ncol
        g1(1:n(ii),1:m(ii),ii) = bdct(coeff_data{ii},[8 8],'inverse');
    end

    %Upsamling of CbCr components
    for ii=2:ncol
        g1(:,:,ii) = imresize(g1(1:n(ii),1:m(ii),ii),[n(1),m(1)]);
    end


%%TGV-decompression

    %%Initialization of iteration-related variables------------------------

        time=cputime;
        g2=zeros(N,M,ncol); %Prepare space for resulting TGV decompresion

        %Primal iterate, extragradient and dual iterate
        x = zeros(N,M,3,ncol); %Primal variable, (:,:,i,j) i'th part of primal variable, j'th color component
        ext = zeros(N,M,3,ncol); %Extragradient
        z = zeros(N,M,5,ncol); %Dual variable, (:,:,i,j) i'th part of dual variable, j'th color component

        %Set new data matrices
        data = zeros(dn(1),dm(1),ncol);
        ldata = zeros(size(data));
        rdata = zeros(size(data));
        d_index = logical(data);
        x_sub = zeros(size(data));

        %Temporary index for iteration steps
        index = logical(data);

        %Adjoint to p
        if p==1
            q = inf;
        elseif p==inf;
            q = 1;
        else
            q = p/(p-1);
        end

        %Dequantize and possibly extend data
        for ii=1:ncol
            offn = 8*(facn(ii));
            offm = 8*(facm(ii));
            
            for i=1:offn:dn(ii)
            for j=1:offm:dm(ii)
                        
                        pn = ((i-1)/facn(ii))+1;
                        pm = ((j-1)/facm(ii))+1;
                        
                        data(    i:i+7, j:j+7 , ii ) = coeff_data{ii}( pn:pn+7 , pm:pm+7 )*sqrt(facn(ii)*facm(ii));
                        d_index(    i:i+7, j:j+7 , ii ) = true;
                        ldata(    i:i+7, j:j+7 , ii ) = data(    i:i+7, j:j+7 , ii ) - 0.5*quant_data{ii}*sqrt(facn(ii)*facm(ii));
                        rdata(    i:i+7, j:j+7 , ii ) = data(    i:i+7, j:j+7 , ii ) + 0.5*quant_data{ii}*sqrt(facn(ii)*facm(ii));
                        
             end
             end
        end
               

        %Setting initial values
        for ii=1:ncol
            x(1:dn(ii),1:dm(ii),1,ii) = bdct(data(1:dn(ii),1:dm(ii),ii),8*[facn(ii),facm(ii)],'inverse');
        end
        %Data is no longer needed, but will be used as temporary storage
        data = zeros(size(data));


        %Upscale data if subsampling operator is used
        if strcmp(subsample_mode,'explicit')
            for ii=1:ncol
                x(:,:,1,ii) = supscale(x(1:dn(ii),1:dm(ii),1,ii),[N,M]);
            end
        end

        %Initialize extragradient
        ext = x;


        tgv=0;
        gap=0;
        %Initialize tgv and gap
        if strcmp(stop_rule,'gap')
            tgv=zeros(update_rate+1,1);
            gap=zeros(update_rate+1,1);

            [tgv(1),gap(1)] = get_tgvjpeg_gap(x,z,alph1,alph0,gamma,p,q,N,M,n,m,subsample_mode,rdata,ldata,d_index);

        elseif (get_gap)
            tgv=zeros(stop_par+1,1);
            gap = zeros(stop_par+1,1);

            [tgv(1),gap(1)] = get_tgvjpeg_gap(x,z,alph1,alph0,gamma,p,q,N,M,n,m,subsample_mode,rdata,ldata,d_index);

        end


    %Start iteration--------------------------------------------------
        k=0;
        cont = true;

        while cont

            %Display iteration and enlarge tgv,gap vectors
            if rem(k+1,update_rate) == 0
                display(['Current iteration: ',num2str(k+1)]);
                if (strcmp(stop_rule,'gap'))
                    tgv = [tgv;zeros(update_rate,1)];
                    gap = [gap;zeros(update_rate,1)];
                end
            end

            %Algorithmic issues----------------------------------------------------

            %set z+=z+sig*A*ext
            z = z + sig*( cat(3, fgrad(ext(:,:,1,:))-ext(:,:,2:3,:),sym_bgrad(ext(:,:,2:3,:)) ) );

            %Project z=(p,q) to {(p,q):|p|<alph1,|q|<alph0}
            z = cat( 3, (z(:,:,1:2,:)./max(1,vec_frob_norm_2(z(:,:,1:2,:))./alph1)), ... %p-component
                        (z(:,:,3:5,:)./max(1,vec_frob_norm_3(z(:,:,3:5,:))./alph0)) );   %q-component


            %Set x+=x-tau*A'(z)
            %size(bdiv(z(:,:,1:2,:)))
            %size(z(:,:,1:2,:)-fdiv(z(:,:,3:5,:)))
            ext=x-tau*cat(3,-bdiv(z(:,:,1:2,:)),-z(:,:,1:2,:)-fdiv(z(:,:,3:5,:)));


            %Projection to data
            %If the subsampling operator us used, we perform
            %   P_U (u) = u + U[ P_D (Su) - Su ]
            if strcmp(subsample_mode,'explicit')
                for ii=1:ncol
                    %Downscale 
                    x_sub(1:dn(ii),1:dm(ii),ii) = sdownscale(ext(:,:,1,ii),[dn(ii),dm(ii)]);
                    %Project to data
                    %BDCT
                    data(1:dn(ii),1:dm(ii),ii) = bdct(x_sub(1:dn(ii),1:dm(ii),ii),[8 8],'forward');
                end
                %Set index and project                
                index = d_index & ( data>rdata );
                data(index) = rdata(index);
                index = d_index & ( data<ldata );
                data(index) = ldata(index);
                for ii=1:ncol
                    %IBDCT + adding of unprojected image
                    x_sub(1:dn(ii),1:dm(ii),ii)  = bdct(data(1:dn(ii),1:dm(ii),ii),[8 8],'inverse') - x_sub(1:dn(ii),1:dm(ii),ii);
                    %Add projection to original image
                    ext(:,:,1,ii) = ext(:,:,1,ii) + supscale(x_sub(1:dn(ii),1:dm(ii),ii),[N,M]);
                end
            %If no subsampling operator is used, we can directly project to
            % (maybe unbounded) data
            elseif strcmp(subsample_mode,'dct')
                for ii=1:ncol
                    %BDCT
                    data(:,:,ii) = bdct(ext(:,:,1,ii),8*[facn(ii),facm(ii)],'forward');
                end
                %Set index and project
                index = d_index & ( data>rdata );
                data(index) = rdata(index);
                index = d_index & ( data<ldata );
                data(index) = ldata(index);
                for ii=1:ncol
                    %IBDCT
                    ext(:,:,1,ii) = bdct(data(:,:,ii),8*[facn(ii),facm(ii)],'inverse');
                end
            else
                error('Wrong subsampling-mode indicator');
            end
            
            


            %Set extragradient
            x=2*ext - x;


            [x,ext] = deal(ext,x); %Swap extragradient and primal variable


            k = k+1; %Update counter


            %Stopping rule evaluation----------------------------------------------
            %Gap:
                if ( strcmp(stop_rule,'gap') || get_gap)

                    %Set new gap
                    [tgv(k+1),gap(k+1)] = get_tgvjpeg_gap(x,z,alph1,alph0,gamma,p,q,N,M,n,m,subsample_mode,rdata,ldata,d_index);

                    if rem(k,100) == 0
                        display(['Gap: ',num2str(gap(k+1))]);
                    end

                    %Test stopping rule
                    if strcmp(stop_rule,'gap')
                        if (gap(k+1)<stop_par)
                            cont = false;
                        end
                    end

                end

            %Iteration:
                if strcmp(stop_rule,'iteration')
                    if k>=stop_par
                        cont = false;
                    end
                end


        end

        
    %Postprocessing--------------------------------------------------------
        %Rescale and save result image
        g2=squeeze(x(:,:,1,:));

        %Crop down tgv and gap
        if ( strcmp(stop_rule,'gap') );
            tgv = tgv(1:k+1);
            gap = gap(1:k+1);
        end

        %Set time
        time = cputime-time;
        display(['Number of iterations:       ',num2str(k)])
        display(['Time(seconds):              ',num2str(time)])


    %Convert to rgb
    if ncol>1
        %Check color space
        if col_info==3
            y = g1;
            rgb = zeros(size(y));
            rgb(:,:,1) = y(:,:,1) + 1.402*y(:,:,3);
            rgb(:,:,2) = y(:,:,1) - 0.34413*y(:,:,2) - 0.71414*y(:,:,3);
            rgb(:,:,3) = y(:,:,1) + 1.772*y(:,:,2);
            g1 = rgb;
            
            y = g2;
            rgb = zeros(size(y));
            rgb(:,:,1) = y(:,:,1) + 1.402*y(:,:,3);
            rgb(:,:,2) = y(:,:,1) - 0.34413*y(:,:,2) - 0.71414*y(:,:,3);
            rgb(:,:,3) = y(:,:,1) + 1.772*y(:,:,2);
            g2 = rgb;
        else
            warning(['Color space conversion for color space ',num2str(col_info),' not implemented. Skipping Conversion.']);
        end
    end

    %Rescale image 
    g1 = (g1+128)./255;
    g2 = (g2+128)./255;
    
    %Crop to [0,1]
    g1(g1>1) = 1;
    g1(g1<0) = 0;
    g2(g2>1) = 1;
    g2(g2<0) = 0;


    %Display images--------------------------------------------------------
        if (display_images)
            %Display both images

            figure('Name','Standard Decompression','NumberTitle','off');
            imshow(g1,[0,255]);

            figure('Name','TGV Decompression','NumberTitle','off');
            imshow(g2,[0,255]);
        end

        %Display gap and tgv
        step = ceil(k/update_rate);

        if ( ( strcmp(stop_rule,'gap') || get_gap ) && plot )

            figure('Name','TGV development','NumberTitle','off');
            plot(tgv(1:step:end));

            figure('Name','Gap development','NumberTitle','off');
            plot(gap(1:step:end));

        end

    %Write parameter-------------------------------
    %Input: kk (iteration number)-------------------
    psz = size(par_list,1);
    for l=1:psz
        par_out{l,1} = par_list{l,1};
        eval(['par{l,2} = ',par_list{l,1},';'])
    end
    par_out{psz+1,1} = 'iteration_nr'; par_out{psz+1,2}=k;
    par_out{psz+2,1} = mfilename;
    %Output: par-----------------------------------
    %----------------------------------------------	

display( '-------------------------------------');
end%End of function
